#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Do everything with a Plotter in context.
"""

from copy import deepcopy
import importlib

import numpy as np

from yampex.annotate import Annotator
from yampex.subplot import Subplotter
from yampex.util import sub


class OptsBase(object):
    def set_loglog(self, yes=True):
        """
        Makes both axes logarithmic, unless called with C{False}.
        """
        self.opts['loglog'] = yes

    def set_semilogx(self, yes=True):
        """
        Makes x-axis logarithmic, unless called with C{False}.
        """
        self.opts['semilogx'] = yes

    def set_semilogy(self, yes=True):
        """
        Makes y-axis logarithmic, unless called with C{False}.
        """
        self.opts['semilogy'] = yes
        
    def set_useLabels(self, yes=True):
        """
        Has annotation labels point to each plot line instead of a legend,
        with text taken from the legend list. Call with C{False} to
        revert to default legend-box behavior.
        """
        self.opts['useLabels'] = yes

    def set_grid(self, yes=True):
        """
        Adds a grid, unless called with C{False}.
        """
        self.opts['grid'] = yes

    def set_timex(self, yes=True):
        """
        Uses intelligent time scaling for the x-axis, unless called with
        C{False}.
        """
        self.opts['timex'] = yes

    def set_firstVectorTop(self):
        """
        Has the first dependent vector (the second argument to the
        L{Plotter} object call) determine the top (maximum) of the
        displayed plot boundary.
        """
        self.opts['firstVectorTop'] = True
        
    def set_bump(self, yes=True):
        """
        If you don't manually set a yscale with a call to L{set_yscale},
        you can call this method to bump the common y-axis upper limit
        to 120% of what Matplotlib decides. Call with C{False} to
        disable the bump.
        """
        self.opts['bump'] = yes

    def set_zeroBottom(self, yes=True):
        """
        Sets the bottom (minimum) of the Y-axis range to zero, unless
        called with C{False}. This is useful for plotting values that
        are never negative and where zero is a meaningful absolute
        minimum.
        """
        self.opts['zeroBottom'] = yes

    def add_marker(self, x):
        """
        Appends the supplied marker style character to the list of markers
        being used. The first and possibly only marker in the list
        applies to the first vector plotted within a given subplot. If
        there is an additional vector being plotted within a given
        subplot (three or more arguments supplied when calling the
        L{Plotter} object, and more than one marker has been added to
        the list, then the second marker in the list is used for that
        second vector plot line. Otherwise, the first (only) marker in
        the list is used for the second plot line as well.
        """
        self.opts['markers'].append(x)

    def add_line(self, x):
        """
        Appends the supplied line style character to the list of line
        styles being used. The first and possibly only line style in
        the list applies to the first vector plotted within a given
        subplot. If there is an additional vector being plotted within
        a given subplot (three or more arguments supplied when calling
        the L{Plotter} object, and more than one line style has been
        added to the list, then the second line style in the list is
        used for that second vector plot line. Otherwise, the first
        (only) line style in the list is used for the second plot line
        as well.
        """
        self.opts['linestyles'].append(x)

    def set_yscale(self, x):
        """
        Manually sets the y-axis scaling.
        """
        self.opts['yscale'] = x

    def set_axvline(self, k):
        """
        Adds a vertical dashed line at the data point with index I{k}.
        """
        self.opts['axvline'] = k
        
    def set_xlabel(self, x):
        """
        Sets the x-axis label. (Ignored if time-x mode has been activated
        with a call to L{set_timex}.)
        """
        self.opts['xlabel'] = x

    def set_ylabel(self, x):
        """
        Sets the y-axis label.
        """
        self.opts['ylabel'] = x

    def add_legend(self, proto, *args):
        """
        Adds the supplied format-substituted text to the list of legend
        entries. You may include a text I{proto}type with
        format-substitution I{args}, or just supply the final text
        string with no further arguments.
        """
        text = sub(proto, *args)
        self.opts['legend'].append(text)

    def clear_legend(self):
        """
        Clears the list of legend entries.
        """
        self.opts['legend'] = []
        
    def add_annotation(self, k, text, kVector=0):
        """
        Adds the supplied I{text} as an annotation at index I{k} of the
        plotted vector. If there is more than one vector being plotted
        within the same subplot, you can have the annotation refer to
        a vector other than the first one by supplying an additional
        integer argument.
        """
        self.opts['annotations'].append((k, text, kVector))

    def clear_annotations(self):
        """
        Clears the list of annotations.
        """
        self.opts['annotations'] = []
        
    def set_title(self, proto, *args):
        """
        Sets a subplot title. You may include a text I{proto}type
        with format-substitution I{args}, or just supply the final
        text string with no further arguments.
        """
        text = sub(proto, *args)
        if self._isSubplot:
            self.opts['title'] = text
        else:
            self._isFigTitle = True
            self.fig.suptitle(text)


class ContextOpts(object):
    def __init__(self, p):
        self.p = p
        self.opts = {}
        self._refresh()

    def _refresh(self):
        self.merged = self.opts.copy()
        for key in self.p.opts:
            if key not in self.merged:
                self.merged[key] = self.p.opts[key]
        
    def __len__(self):
        self._refresh()
        return len(self.merged)

    def __getitem__(self, name):
        self._refresh()
        return self.merged[name]

    def __setitem__(self, name, value):
        self.opts[name] = value

    def __iter__(self):
        return iter(self.merged)

    def get(self, name, value, default):
        self._refresh()
        return self.merged.get(name, value, default)

    def copy(self):
        self._refresh()
        return self.merged.copy()
    

class SpecialAx(object):
    """
    I pretend to be an L{matplotlib.Axes} object except that I
    intercept plotting calls and rescale the independent vector first.
    """
    _plotterNames = {
        'plot', 'loglog',
        'semilogx', 'semilogy', 'scatter', 'step', 'bar', 'stem'}
    
    def __init__(self, ax, opts, V):
        self.ax = ax
        self.opts = opts
        self.V = V

    def __getattr__(self, name):
        def wrapper(*args, **kw):
            args = list(args)
            if self.V is not None:
                for k, arg in enumerate(args[:2]):
                    if isinstance(arg, str) and arg in self.V:
                        args[k] = self.V[arg]
            if xscale:
                args[0] = args[0] * xscale
            return x(*args, **kw)

        x = getattr(self.ax, name)
        if name not in self._plotterNames:
            return x
        xscale = self.opts.get('xscale', None)
        if xscale or self.V is not None:
            return wrapper
        return x

    
class Plotter(OptsBase):
    """
    I provide a Figure with one or more time-vector and XY subplots of
    Numpy vectors using Matplotlib.

    If you supply an object other than a list or tuple that houses
    vectors and provides access to them by name as items (i.e., a dict
    or a Vectors object from the I{pingspice} package) as a third
    argument, I will convert vector names to vectors for you in each
    plotting call.

    Any keywords you supply to the constructor are used to C{set\_X}
    the axes for all subplots. For example, C{yticks=[1,2,3]} results
    in a C{set\_yticks([1,2,3])} command to the C{axes} object for all
    subplots.
    """
    figSize = (10.0, 7.0)
    colors = ['b', 'g', 'r', '#40C0C0', '#C0C040', '#C040C0', '#8080FF']
    timeScales = [
        (1E-9, "nanoseconds"),
        (1E-6, "microseconds"),
        (1E-3, "milliseconds"),
    ]

    _opts = {
        'marker':               '',
        'linestyle':            '-',
        'markers':              [],
        'linestyles':           [],
        'minorTicks':           [],
        'grid':                 False,
        'firstVectorTop':       False,
        'loglog':               False,
        'semilogx':             False,
        'semilogy':             False,
        'legend':               [],
        'annotations':          [],
        'xscale':               None,
        'yscale':               None,
        'useLabels':            False,
        'axvline':              None,
        'bump':                 False,
        'timex':                False,
        'xlabel':               "",
        'ylabel':               "",
        'title':                "",
        'zeroBottom':           False,
    }
    _settings = {'title', 'xlabel', 'ylabel'}
    _plotters = []
    
    def __init__(self, N, *args, **kw):
        args = list(args)
        if args:
            Nx = N
            Ny = args.pop(0)
        else:
            Nx = self.N_cols(N)
            Ny = int(np.ceil(float(N)/Nx))
        self.V = args[0] if args else None
        self._plotters.append(self)
        self.opts = deepcopy(self._opts)
        self.filePath = kw.pop('filePath', None)
        self.plt = importlib.import_module("matplotlib.pyplot")
        figSize = list(self.figSize)
        if 'width' in kw:
            figSize[0] = kw.pop('width')
        if 'height' in kw:
            figSize[1] = kw.pop('height')
        self.fig = self.plt.figure(figsize=figSize)
        self.sp = Subplotter(self, Nx, Ny)
        self.annotators = {}
        self.kw = kw
        self._isFigTitle = False
        self._isSubplot = False
        
    def __getattr__(self, name):
        return self.opts[name]
        
    def __enter__(self):
        self.sp.setup()
        self._isSubplot = True
        self.originalOpts = self.opts.copy()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sp.postOp()
        self._isSubplot = False
        self.opts = self.originalOpts

    def N_cols(self, N):
        if N > 6:
            return 3
        if N > 3:
            return 2
        return 1

    def show(self, windowTitle=None, subcall=False):
        if not subcall and len(self._plotters) > 1:
            for p in self._plotters:
                p.show(windowTitle, subcall=True)
            self.plt.show()
            return
        self.fig.tight_layout()
        if self._isFigTitle:
            self.fig.subplots_adjust(top=0.93)
        self.plt.draw()
        for annotator in self.annotators.values():
            annotator.update()
        if self.filePath is None:
            self.plt.draw()
            if windowTitle:
                self.fig.canvas.set_window_title(windowTitle)
            if not subcall:
                self.plt.show()
            return
        fh = open(self.filePath, 'wb+')
        self.fig.savefig(fh, format='png')
        self.plt.close()
        fh.close()

    def getColor(self, k):
        return self.colors[k % len(self.colors)]
        
    def xBounds(self, *args, **kw):
        self.sp.xBounds(*args, **kw)
    
    def yBounds(self, *args, **kw):
        self.sp.yBounds(*args, **kw)

    def scaleTime(self, vectors):
        V0 = vectors[0]
        if not self.timex:
            return V0
        T_max = vectors[0].max()
        for mult, name in self.timeScales:
            if T_max < 1000*mult:
                break
        else:
            self.opts['xlabel'] = "seconds"
            return V0
        self.opts['xlabel'] = name
        self.opts['xscale'] = 1.0 / mult
        return V0 / mult 

    def parseArgs(self, args):
        vectors = []
        if not isinstance(args[0], (list, tuple, np.ndarray)):
            V = args[0]
            names = args[1:]
            for name in names:
                vectors.append(V[name])
            return vectors, names, V
        if self.V is not None:
            names = []
            for arg in args:
                if isinstance(arg, str) and arg in self.V:
                    names.append(arg)
                    vectors.append(self.V[arg])
                else:
                    names.append(None)
                    vectors.append(arg)
            if None in names:
                names = None
            return vectors, names, self.V
        return args, None, None

    def addLegend(thisVector, thisLegend):
        # Put an annotation at the right-most point where the
        # value is at least 99.9% of the vector maximum
        if max(thisVector) > 0:
            m999 = np.greater(thisVector, 0.999*max(thisVector))
        else:
            m999 = np.less(thisVector, 0.999*min(thisVector))
        k = max(np.nonzero(m999)[0])
        self.annotations.append((kVector, k, thisLegend))
    
    def __call__(self, *args, **kw):
        """
        Plots the second supplied vector (and any further ones) versus the
        first in the next subplot. If you supply an object that houses
        vectors and provides access to them as items, e.g.,
        C{analysis.sim.Vectors} of the I{pingspice} package, as the
        first argument, you can supply vector names instead of the
        vectors themselves.
        
        Options are set via the methods in L{OptsBase}, including a
        I{title}, a list of plot I{markers} and I{linestyles}, and a
        list of I{legend} entries for the plots with those
        keywords. I{legend} can also be set C{True} to use the default
        Matplotlib legend.

        Set I{useLabels} to C{True} to have annotation labels pointing
        to each plot line instead of a legend, with text taken from
        the legend list. (Works best in interactive apps.)

        Add other text annotations by the sample index of the point
        within the independent (first) and dependent vectors, the
        text, and the vector index. The vector index starts with 0 for
        the second vector argument and first dependent variable
        vector.

        Rescale all vectors after the first dependent one, relative to
        that first dependent one, by supplying a I{yscale}. This will
        result in two different twinned x-axes (one for the first
        dependent vector and one for everybody else) and a different
        y-axis label on the right.

        If you don't provide any such yScale, you can set I{bump} to
        C{True} to bump the common y-axis upper limit to 120% of what
        Matplotlib decides.

        If your x-axis is for time with units in seconds, you can set
        I{timex} to C{True} and the x values will be in set to the
        most sensible units, e.g., nanoseconds for values < 1E-6. Any
        'xlabel' keyword is disregarded in such case because the x
        label is set automatically.
        
        Any keywords you supply to this call are used to C{set\_X} the
        axes, e.g., C{ylabel="foo"} results in a C{set\_ylabel("foo")}
        command to the C{axes} object, for this subplot only.

        If you want to do everything with the next subplot on your own
        and only want a reference to its C{Axes} object, just call
        this with no args. You can still supply keywords to do
        C{set\_X} stuff if you wish.
        """
        def getLast(obj, k):
            return obj[k] if k < len(obj) else obj[-1]

        def plotVector(k, vector, ax):
            if self.loglog: plotter = ax.loglog
            elif self.semilogx: plotter = ax.semilogx
            elif self.semilogy: plotter = ax.semilogy
            else: plotter = ax.plot 
            kw = {'scaley': not self.firstVectorTop}
            kw['marker'] = getLast(self.markers, k) \
                if self.markers else self.marker
            kw['linestyle'] = getLast(self.linestyles, k) \
                if self.linestyles else self.linestyle
            if kw['marker'] in (',', '.'): kw['linestyle'] = ''
            if kw['linestyle'] == '-': kw['linewidth'] = 2
            color = self.getColor(k)
            kw['color'] = color
            # Here is where the plotting actually happens
            lineInfo[0].extend(plotter(firstVector, vector, **kw))
            if yscale:
                ax.tick_params('y', colors=color)
                if k == 0:
                    ax.set_ylim(bottom=0)
                else: ax.set_ylim(0, axFirst.get_ylim()[1]/yscale)
            if isinstance(self.legend, bool):
                if self.legend and names:
                    legend = names[k+1]
                else:
                    return
            elif k < len(self.legend):
                legend = self.legend[k]
            else:
                return
            lineInfo[1].append(legend)
            if yscale: ax.set_ylabel(legend, color=color)
            if self.useLabels: self.addLegend(vector, legend)
        
        def adjustPlots(yscale, axvline):
            if self.bump and yscale is None:
                self.yBounds(ax, bump=True, zeroBottom=self.zeroBottom)
            elif self.firstVectorTop:
                self.yBounds(ax, Ymax=yMax, zeroBottom=self.zeroBottom)
            if isinstance(axvline, int) and abs(axvline) < len(firstVector):
                axFirst.axvline(
                    x=firstVector[axvline], linestyle='--', color="#404040")
            if self.legend and not self.annotations and not self.useLabels:
                axFirst.legend(*lineInfo, **{'loc': "best"})

        def doSettings(kw):
            for name in self.kw:
                kw.setdefault(name, self.kw[name])
            for name in self._settings:
                if self.opts[name]:
                    kw.setdefault(name, self.opts[name])
            for name in kw:
                self.sp.set_(name, kw[name])

        def doAnnotations(yscale):
            self.plt.draw()
            annotator = self.annotators[axFirst] = Annotator(
                axList, [firstVector]+list(vectors[1:]))
            for k, text, kVector in self.annotations:
                x = firstVector[k]
                y = vectors[kVector+1][k]
                kAxis = 0 if yscale is None or kVector == 0 else 1
                annotator(kAxis, x, y, text)
            annotator.update()

        ax = axFirst = self.sp[None]
        if not args:
            doSettings(kw)
            return ax
        lineInfo = [[],[]]
        yscale = self.yscale
        vectors, names, V = self.parseArgs(args)
        firstVector = self.scaleTime(vectors)
        N = len(firstVector)
        yMax = -np.inf
        for kVector, thisVector in enumerate(vectors[1:]):
            thisMax = thisVector.max()
            if thisMax > yMax:
                yMax = thisMax
            if yscale and kVector == 1:
                ax = self.sp.twinx()
            plotVector(kVector, thisVector, ax)
        adjustPlots(yscale, self.axvline)
        doSettings(kw)
        axList = self.sp.getTwins()
        if self.annotations:
            doAnnotations(yscale)
        axList = [SpecialAx(ax, self.opts, V) for ax in axList]
        if len(axList) == 1:
            return axList[0]
        return axList



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Simple subplotting.
"""

class Subplotter(object):
    def __init__(self, plotter, Nx, Ny):
        self.p = plotter
        N = Nx * Ny
        self.axes = []
        self.twins = {}
        self.kLast = None
        for k in range(N):
            ax = plotter.fig.add_subplot(Ny, Nx, k+1)
            self.axes.append(ax)

    def setup(self):
        self.kLast = None
        for ax in self.axes:
            ax.clear()
            for axTwin in self.getTwins(ax, mplRoster=True):
                axTwin.clear()
            if ax in self.twins:
                self.twins[ax][0] = 0

    def postOp(self):
        ax = self.ax
        if ax is None: return
        if self.kLast >= len(self.p.minorTicks) \
           or self.p.minorTicks[self.kLast]:
            ax.minorticks_on()
        if self.p.grid:
            ax.grid(True, which='major')
            
    @property
    def ax(self):
        if self.kLast is not None and self.kLast < len(self.axes):
            return self.axes[self.kLast]
        
    def __len__(self):
        return len(self.axes)

    def __getitem__(self, k):
        if self.kLast is None:
            # First default is first element.
            if k is None: k = 0
        else:
            # An element index previously has been specified: Do
            # post-op on it
            self.postOp()
            # ...and default to the next element
            if k is None: k = self.kLast+1
        self.kLast = k
        if k < len(self.axes):
            return self.axes[k]

    def set_(self, what, name, **kw):
        f = getattr(self.ax, "set_{}".format(what))
        f(name, **kw)
        
    def xBounds(self, ax=None, left=None, right=None):
        if ax is None: ax = self.ax
        if left and left < ax.get_xlim()[0]:
            ax.set_xlim(left=left)
        if right and right > ax.get_xlim()[1]:
            ax.set_xlim(right=right)
            
    def yBounds(self, ax=None, Ymax=None, bump=False, zeroBottom=False):
        """
        If I{bump} is C{True} or I{Ymax} is a value greater than 95% of
        the largest y-value currently plotted, increases the upper
        bound by 20%. If I{zeroBottom} is set C{True}, anchors the
        bottom to zero.
        """
        def top(ax):
            return ax.get_ylim()[1]
        
        def updateYlim(ax):
            kw = {}
            if zeroBottom: kw['bottom'] = 0
            if bump: kw['top'] = 1.2*top(ax)
            ax.set_ylim(**kw)
        
        if ax is None: ax = self.ax
        if not bump:
            if Ymax is None:
                lines = ax.get_lines()
                if not lines:
                    return
                Ymax = max([max(x.get_ydata(orig=True)) for x in lines])
            if Ymax > 0.95 * top(ax):
                bump = True
        if zeroBottom or bump:
            for axThis in self.getTwins(ax):
                updateYlim(axThis)

    def twinx(self, ax=None):
        """
        Creates and returns a twin of the last (or supplied) axes object,
        maintaining a list of the twinned axes in the order they were
        created along with an index of the latest-created twin (for
        axes re-use).
        """
        if ax is None: ax = self.ax
        # Shouldn't need redundant storage from what MatplotLib does,
        # but whatever
        if ax in self.twins:
            k, twinList = self.twins[ax]
        else:
            k = 0
            twinList = []
            self.twins[ax] = [k, twinList]
        if k >= len(twinList):
            twinList.append(ax.twinx())
        twin = twinList[k]
        self.twins[ax][0] = k+1
        return twin

    def getTwins(self, ax=None, mplRoster=False):
        """
        Returns a list of the twins for the last (or supplied) axes
        object, starting with the axes object itself and then the
        twins in the order they were created.
        """
        if ax is None: ax = self.ax
        if mplRoster:
            # API hint to Matplotlib devs: Don't unecessarily nest lists!
            groupAsList = list(ax.get_shared_x_axes())
            return groupAsList[0] if groupAsList else []
        twins = [ax]
        if ax in self.twins:
            k, twinList = self.twins[ax]
            twins.extend(twinList[:k])
        return twins



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
B{Y}et B{A}nother B{M}atB{p}lotlib B{Ex}tension, with simplified
subplotting.
"""

from plot import Plotter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import numpy as np
from yampex import Plotter

funcNames = ('sin', 'cos')
X = np.linspace(0, 4*np.pi, 200)
pt = Plotter(1, 2, width=7, height=5)
pt.set_title("Sin and Cosine")
pt.set_xlabel("X"); pt.set_grid()
pt.add_annotation(199, "Last")
with pt as p:
    for funcName in funcNames:
        Y = getattr(np, funcName)(X)
        p.set_ylabel("{}(X)".format(funcName))
        k = 0 if funcName == 'sin' else 75
        for text in ("Pos ZC", "Max", "Neg ZC", "Min"):
            p.add_annotation(k, text)
            k += 25
        p.set_axvline(k)
        p(X, Y)
pt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.


NAME = "yampex"


### Imports and support
from setuptools import setup

### Define requirements
required = ['numpy', 'matplotlib',]


### Define setup options
kw = {'version':'0.8.1',
      'license':'Apache License (2.0)',
      'platforms':'OS Independent',

      'url':"http://edsuom.com/{}.html".format(NAME),
      'author':"Edwin A. Suominen",
      'author_email':"foss@edsuom.com",
      'maintainer':'Edwin A. Suominen',
      'maintainer_email':"foss@edsuom.com",
      
      'install_requires':required,
      'packages':['yampex',],
      
      'zip_safe':True,
}

kw['keywords'] = [
    'matplotlib', 'numpy',
    'extension', 'subplots', 'annotations', 'plotting',
]


kw['classifiers'] = [
    'Development Status :: 4 - Beta',

    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Framework :: Twisted',

    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Scientific/Engineering :: Visualization',
]


kw['description'] = " ".join("""
Yet Another Matplotlib Extension, with simplified subplotting.
""".split("\n"))

kw['long_description'] = """
yampex makes Matplotlib_ easier to use, especially with subplots. You
simply construct a Plotter_ object with the number of subplot rows
and columns you want, and do a context call on it to get a version of
the object that's all set up to do your subplots.

There's a quick example on the project page_ at edsuom.com.

.. _Matplotlib: https://matplotlib.org/

.. _Plotter: http://edsuom.com/yampex/yampex.plot.Plotter.html

.. _page: http://edsuom.com/yampex.html

"""

### Finally, run the setup
setup(name=NAME, **kw)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# yampex:
# Yet Another Matplotlib Extension
#
# Copyright (C) 2017-2018 by Edwin A. Suominen,
# http://edsuom.com/yampex
#
# See edsuom.com for API documentation as well as information about
# Ed's background and other projects, software and otherwise.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS
# IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Smart plot annotations.
"""

from copy import copy

import numpy as np
import matplotlib.patches as patches

from yampex.util import sub


class Sizer(object):
    """
    I try to provide accurate sizing of annotations.

    Burned up a lot of CPU cycles in the __call__ method, which is
    called many thousands of times. Added use of I{legitDims} list,
    which speeds things up and doesn't seem to incur any problems with
    mis-sizing of evaluation rectangles.
    """
    maxTries = 3
    bogusHeightThreshold = 5 # pixels

    __slots__ = ['dims', 'legitDims']
    
    def __init__(self):
        self.dims = {}
        self.legitDims = []

    def area(self, dims):
        return dims[0] * dims[1]
        
    def __call__(self, ann, tryCount=0):
        text = ann.get_text()
        if text in self.legitDims:
            # Taking a chance that a non-bogus size previously
            # computed can be relied upon. Seems to work, and saves
            # considerable processing time.
            return self.dims[text]
        bb = ann.get_bbox_patch().get_extents()
        theseDims = (bb.width, bb.height)
        if theseDims[1] < self.bogusHeightThreshold and \
           tryCount < self.maxTries:
            ann.draw(ann.axes.figure.canvas.get_renderer())
            return self(ann, tryCount+1)
        prevDims = self.dims.get(text, (0,0))
        if self.area(theseDims) > self.area(prevDims):
            self.dims[text] = theseDims
            self.legitDims.append(text)
        return self.dims[text]


class Rectangle(object):
    """
    I am a rectangular plot region where an annotation is or might go.
    """
    colors = ['#ff8080', '#c0c080', '#80c0c0', '#c080c0']
    drawnBoxes = []
    
    def __init__(self, ann, relpos, x01, y01):
        self.ann = ann
        self.relpos = relpos
        self.xy = ann.axes.transData.transform(ann.xy)
        self.x0, self.x1 = x01
        self.y0, self.y1 = y01

    def __repr__(self):
        return sub(
            "({:.0f}, {:.0f}) <--- [{:.0f},{:.0f} {:.0f},{:.0f}]",
            self.xy[0], self.xy[1],
            self.x0, self.y0, self.x1, self.y1)

    def xOverlap(self, other):
        if self.x1 < other.x0:
            # My annotation is fully to the left of this one
            return False
        if self.x0 > other.x1:
            # Mine is fully to the right of it
            return False
        return True
        
    def yOverlap(self, other):
        if self.y1 < other.y0:
            # My annotation is fully below this one
            return False
        if self.y0 > other.y1:
            # Mine is fully above it
            return False
        return True

    def overlap(self, other):
        return self.xOverlap(other) and self.yOverlap(other)

    def arrowOverlap(self, other):
        def zBetweenAB(a, b, z):
            """Is z between a and b? (Make sure a < b)"""
            return z > a and z < b
        
        x, y = self.xy
        if self.relpos[0] == 0.5:
            #import pdb; pdb.set_trace()
            # Vertical arrow
            if x < other.x0 or x > other.x1:
                return False
            # Arrow between other's vertical sides
            if self.y1 < other.y0:
                # Other is above me...
                if y < self.y0 or zBetweenAB(self.y1, other.y0, y):
                    # ...so no overlap if arrow points down or is
                    # between us
                    return False
            elif y > self.y1 or zBetweenAB(other.y1, self.y0, y):
                # Other is below me, so no overlap if arrow points up
                # or is in between us
                return False
            # There must be overlap
            return True
        if self.relpos[1] == 0.5:
            # Horizontal arrow
            if y < other.y0 or y > other.y1:
                return False
            # Arrow between other's horizontal sides
            if self.x1 < other.x0:
                # Other is to my right...
                if x < self.x0 or zBetweenAB(self.x1, other.x0, x):
                    # ...so no overlap if arrow points to the left or
                    # is between us
                    return False
            elif x > self.x1 or zBetweenAB(other.x1, self.x0, x):
                # Other is to my left, so no overlap if arrow points
                # to the right or is in between us
                return False
            # There must be overlap
            return True
        # TODO: Angled arrow, not (yet) tested
        return False

    def draw(self):
        if self.ann not in self.drawnBoxes:
            self.drawnBoxes.append(self.ann)
        color = self.colors[
            self.drawnBoxes.index(self.ann) % len(self.colors)]
        r = patches.Rectangle(
            [self.x0, self.y0], self.x1-self.x0, self.y1-self.y0,
            color=color, fill=False)
        if not hasattr(self, 'fig'):
            self.fig = self.ann.axes.get_figure()
        self.fig.patches.append(r)
        self.fig.draw_artist(r)

    @classmethod
    def clear(cls, fig):
        cls.drawnBoxes = []
        patches = fig.patches
        for r in patches:
            patches.remove(r)


class Positioner(object):
    """
    All data is saved and computations done in pixel units.
    """
    relpos = [
        (0.0, 0.0),     # NE
        (0.0, 0.5),     # E
        (0.0, 1.0),     # SE
        (0.5, 1.0),     # S
        (1.0, 1.0),     # SW
        (1.0, 0.5),     # W
        (1.0, 0.0),     # NW
        (0.5, 0.0),     # N
    ]
    rectanglePadding = 2
    # I hate fudge factors, but rectangle analysis region is shifted
    # too far without this
    ffShift = 5
    
    def __init__(self, sizer, axList, vectors):
        self.sizer = sizer
        self.axList = axList
        self.xyLists = {}
        self.vectors = vectors
        self.annList = []
        self.prevData(vectors)

    @property
    def relpos(self):
        if not hasattr(self, 'r'):
            return
        return self.r.relpos
    
    def __iter__(self):
        for ann in copy(self.annList):
            if ann in self.annList:
                yield ann
    
    def add(self, ann):
        self.annList.append(ann)
        
    def remove(self, ann):
        if ann in self.annList:
            self.annList.remove(ann)
    
    def dataToPixels(self, ann=None, kAxes=None, ax=None, xyData=None):
        if ax is None:
            ax = self.axList[kAxes] if ann is None else ann.axes
        if xyData is None:
            if kAxes is None:
                xy = ann.xy
            else:
                xy = np.column_stack(
                    (self.vectors[0], self.vectors[kAxes+1]))
        else:
            xy = xyData
        XY = ax.transData.transform(xy)
        if len(XY.shape) > 1:
            return [XY[:,k] for k in (0,1)]
        return XY
                
    def rectangle(self, ann, dx, dy):
        """
        Sets up a new overlap analysis with the supplied annotation and
        proposed x and y offset (pixel units) from its current location.

        The offset is adjusted so that positive x or y offset puts the
        left or lower corner at that point, negative x or y offset
        puts the right or upper corner there, and zero x or y offset
        puts the middle of the edge there.

        Returns a L{Rectangle} with the four points
        """
        def shift(x_y, dx_y, dim):
            if dx_y < 0:
                relpos.append(1.0)
                x_y1 = x_y + dx_y + self.ffShift
                x_y0 = x_y1 - dim
            elif dx_y > 0:
                relpos.append(0.0)
                x_y0 = x_y + dx_y - self.ffShift
                x_y1 = x_y0 + dim
            else:
                relpos.append(0.5)
                x_y0 = x_y - 0.5*dim
                x_y1 = x_y + 0.5*dim
            return x_y0 - self.rectanglePadding, x_y1 + self.rectanglePadding
        
        relpos = []
        x, y = self.dataToPixels(ann=ann)
        width, height = self.sizer(ann)
        x01 = shift(x, dx, width)
        y01 = shift(y, dy, height)
        r = Rectangle(ann, relpos, x01, y01)
        return r

    def setup(self, ann, dx, dy):
        """
        Sets up a new overlap analysis with the supplied annotation and
        proposed x and y offset (pixel units) from its current location.
        """
        self.dx = dx
        self.ann = ann
        self.r = self.rectangle(ann, dx, dy)

    def liveData(self, kAxes, *xy):
        xyLists = self.xyLists.setdefault(kAxes, [[], []])
        if xy:
            x, y = xy
            k = np.searchsorted(xyLists[0], x)
            xyLists[0].insert(k, x)
            xyLists[1].insert(k, y)
        return np.column_stack(xyLists)

    def prevData(self, XY=None):
        if XY is None:
            return getattr(self, '_prevData', None)
        self._prevData = XY
    
    def datarator(self):
        """
        Yields X and Y transformed to pixels with current axes display
        """
        N = len(self.axList)
        for kAxes in range(N):
            yield self.dataToPixels(kAxes=kAxes)
            xyData = self.liveData(kAxes)
            if xyData.shape[0]:
                yield self.dataToPixels(kAxes=kAxes, xyData=xyData)
        xyData = self.prevData()
        if xyData is not None:
            yield self.dataToPixels(kAxes=kAxes, xyData=xyData)
    
    def with_data(self):
        """
        Returns score for overlapping with X,Y data for all
        axes. Increases with number of plot lines overlapped.
        """
        def sliceSpanningRectangle():
            k0 = np.searchsorted(X, self.r.x0)
            k1 = np.searchsorted(X, self.r.x1) + 1
            return slice(k0, k1)
        
        s = None
        score = 0.0
        for X, Y in self.datarator():
            # Get slice of Y spanning same X interval as annotation
            newX = False
            if s is None:
                # First data set
                Xmm = X.min(), X.max()
                s = sliceSpanningRectangle()
            else:
                thisXmm = X.min(), X.max()
                if thisXmm != Xmm:
                    # Data set with different X interval from the
                    # previous one
                    Xmm = thisXmm
                    s = sliceSpanningRectangle()
            if np.all(np.less(Y[s], self.r.y0)): continue
            if np.all(np.greater(Y[s], self.r.y1)): continue
            score += 1.5
            if score >= self.mustBeat: break
        return score

    def with_boundary(self):
        """
        Returns score for overlapping with (or going beyond) axis
        boundary, and twice as bad a score for such a conflict with
        the figure boundary.
        """
        score = 0.0
        ax = self.ann.axes
        for k, obj in enumerate((ax, ax.figure)):
            points = obj.get_window_extent().get_points()
            x0, y0 = points[0]
            x1, y1 = points[1]
            if self.r.x0 < x0 or \
               self.r.x1 > x1 or \
               self.r.y0 < y0 or \
               self.r.y1 > y1:
                score += 3*(k+1)
            else:
                return score
        return score

    def with_others(self):
        """
        Returns score for overlapping with any other annotation.
        """
        score = 0.0
        for other in self:
            if other is self.ann:
                continue
            tr = self.rectangle(other, *other.xyann)
            if self.r.overlap(tr):
                return 4.0
            # Return somewhat less bad score if there is arrow overlap
            if self.r.arrowOverlap(tr):
                score += 2.0
            # TODO: Account for angled arrows, too
            if score >= self.mustBeat:
                break
        return score

    def __call__(self, ann, offset, radius, mustBeat=1E9):
        offset = [radius*x for x in offset]
        self.setup(ann, *offset)
        self.mustBeat = mustBeat
        #self.r.draw() # DEBUG
        score = self.with_boundary()
        if score < mustBeat:
            score += self.with_others()
        if score < mustBeat:
            score += self.with_data()
        return radius*score
    
    
class Annotator(object):
    """
    I manage the annotations for all axes in a subplot. Construct me
    with a list of the axes and call the instance to add each new
    annotation.
    """
    rp = 28.3
    offsets = [
        (+20, +20),     # NE
        (+rp,   0),     # E
        (+20, -20),     # SE
        (  0, -rp),     # S
        (-20, -20),     # SW
        (-rp,   0),     # W
        (-20, +20),     # NW
        (  0, +rp),     # N
    ]
    moreOffsets = [
        (+11, +26),     # NNE
        (+26, +11),     # ENE
        (+26, -11),     # ESE
        (+11, -26),     # SSE
        (-11, -26),     # SSW
        (-26, -11),     # WSW
        (-26, +11),     # WNW
        (-11, +26),     # NNW
    ]
    radii = [1.0, 2.0, 4.0, 8.0, 12.0, 16.0]

    fontSize = 12 # points
    fontWeight = 'normal'
    arrowprops = {
        'facecolor':            "#800000",
        'edgecolor':            "#800000",
        'arrowstyle':           "-|>",
        'connectionstyle':      "arc3",
        'relpos':               (0.5, 0.5),
    }
    boxprops = {
        'boxstyle':             "round",
        'facecolor':            "white",
        'edgecolor':            "#800000",
        'lw':                   2,
        'alpha':                0.8,
    }
    maxDepth = 10

    # PROFILE
    #from cProfile import Profile
    #P = Profile()
    
    def __init__(self, axList, vectors, **kw):
        self.axList = axList
        self.sizer = Sizer()
        self.p = Positioner(self.sizer, axList, vectors)
        for name in kw:
            setattr(self, name, kw[name])
    
    def offseterator(self, radius):
        for k, offsets in enumerate(self.offsets):
            yield offsets
            if radius > 2:
                yield self.moreOffsets[k]
        
    def scaledOffset(self, radius, offset):
        return [radius*x for x in offset]
        
    def textAlignment(self, relpos):
        if relpos[0] == 0:
            ha = "left"
        elif relpos[0] == 1:
            ha = "right"
        else:
            ha = "center"
        if relpos[1] == 0:
            va = "bottom"
        elif relpos[1] == 1:
            va = "top"
        else:
            va = "center"
        return ha, va

    def evaluate(self, ann, depth=0):
        def setBest():
            best['score'] = score
            best['offset'] = self.scaledOffset(radius, offset)
            best['relpos'] = self.p.relpos

        def realDifference(a, b):
            for k in (0, 1):
                if abs(a[k] - b[k]) > 0.5: return True
            return False

        def radiusPenalty(radius):
            if radius > 2:
                return 0.2*np.sqrt(radius)
            return 0

        best = {'score': 1E9}
        for k, radius in enumerate(self.radii):
            for offset in self.offseterator(radius):
                score = self.p(ann, offset, radius, mustBeat=best['score'])
                if score < best['score']: setBest()
                if score == 0: break
            if best['score'] == 0: break
            best['score'] += radiusPenalty(radius)
            if k+1 < len(self.radii):
                nextRadius = self.radii[k+1]
                if best['score'] < radiusPenalty(nextRadius):
                    # Even a clear spot would be scored worse at next
                    # higher radius, so we're done
                    break
            
        offset = best['offset']
        if realDifference(offset, ann.xyann):
            return self.replace(
                ann, offset=offset, relpos=best['relpos'], depth=depth)
        return ann
    
    def update(self, *args, **kw):
        depth = kw.pop('depth', 0) + 1
        if depth > self.maxDepth:
            return False
        updated = False
        for ann in self.p:
            result = self.evaluate(ann, depth)
            if result != ann:
                updated = True
        return updated
    
    def add(self, ax, text, xy, offset=None, relpos=None, depth=0):
        if offset is None: offset = (0,0)
        arrowprops = self.arrowprops.copy()
        if relpos:
            arrowprops['relpos'] = relpos
            ha, va = self.textAlignment(relpos)
        else:
            ha, va = ['center']*2
        ann = ax.annotate(
            text, xy=xy, xytext=offset,
            textcoords="offset pixels",
            ha=ha, va=va, size=self.fontSize, weight=self.fontWeight,
            arrowprops=arrowprops, bbox=self.boxprops, zorder=100)
        self.p.add(ann)
        ann.draw(ax.figure.canvas.get_renderer())
        # PROFILE
        # ---------------------------------------------
        #self.P.runcall(self.update, depth=depth)
        self.update(depth=depth)
        # ---------------------------------------------
        return ann

    def remove(self, ann, ax=None):
        if ax is None: ax = ann.axes
        for thisAnn in ax.findobj(match=type(ann)):
            if thisAnn.get_text() == ann.get_text():
                thisAnn.remove()
                self.p.remove(thisAnn)
        self.p.remove(ann)
    
    def replace(self, ann, **kw):
        ax = kw.get('ax', None)
        if ax is None: ax = ann.axes
        xy = kw.get('xy', None)
        if xy is None: xy = ann.xy
        text = kw.get('text', None)
        if text is None: text = ann.get_text()
        self.remove(ann, ax)
        return self.add(
            ax, text, xy,
            kw.get('offset', None), kw.get('relpos', None), kw.get('depth', 0))
    
    def __call__(self, kAxis, x, y, text):
        ax = self.axList[kAxis]
        ann = self.add(ax, text, (x, y))
        return ann




        
